from sqlalchemy.dialects.oracle.base import OracleDialect
from sqlalchemy.dialects.oracle.cx_oracle import _OracleBINARY_DOUBLE
from sqlalchemy.dialects.oracle.cx_oracle import _OracleBINARY_FLOAT
from sqlalchemy.dialects.oracle.cx_oracle import _OracleInteger
from waddle import ParamBunch

from .base import BaseConnector
from sqlalchemy.dialects.oracle import cx_oracle
from sqlalchemy.dialects import registry
from . import oracle_db_wrapper



class OracleDbDialect(cx_oracle.OracleDialect_cx_oracle):
    def __init__(
            self,
            auto_convert_lobs=True,
            coerce_to_unicode=True,
            coerce_to_decimal=True,
            arraysize=50,
            encoding_errors=None,
            threaded=None,
            **kwargs):
        import oracledb
        OracleDialect.__init__(self, **kwargs)
        self.arraysize = arraysize
        self.encoding_errors = encoding_errors
        if threaded is not None:
            self._cx_oracle_threaded = threaded
        self.auto_convert_lobs = auto_convert_lobs
        self.coerce_to_unicode = coerce_to_unicode
        self.coerce_to_decimal = coerce_to_decimal
        self.dbapi = self.__class__.dbapi()
        if self._use_nchar_for_unicode:
            self.colspecs = self.colspecs.copy()
            self.colspecs[sqltypes.Unicode] = _OracleUnicodeStringNCHAR
            self.colspecs[sqltypes.UnicodeText] = _OracleUnicodeTextNCLOB

        cx_Oracle = self.dbapi

        self._include_setinputsizes = {
            oracledb.DATETIME,
            oracledb.NCLOB,
            oracledb.CLOB,
            oracledb.LOB,
            oracledb.NCHAR,
            oracledb.FIXED_NCHAR,
            oracledb.BLOB,
            oracledb.FIXED_CHAR,
            oracledb.TIMESTAMP,
            _OracleInteger,
            _OracleBINARY_FLOAT,
            _OracleBINARY_DOUBLE,
        }

        self._paramval = lambda value: value.getvalue()

        # https://github.com/oracle/python-cx_Oracle/issues/176#issuecomment-386821291
        # https://github.com/oracle/python-cx_Oracle/issues/224
        self._values_are_lists = False
        self._returningval = self._paramval

        self._is_cx_oracle_6 = True

    @classmethod
    def dbapi(cls):
        import oracledb
        return oracledb

    def do_execute(self, cursor, statement, parameters, context=None):
        if isinstance(statement, bytes):
            statement = statement.decode('utf-8')
        super().do_execute(cursor, statement, parameters, context)


registry.register('oracledb', __name__, 'OracleDbDialect')


class OracleDb(BaseConnector):
    default_port = 1521
    engine = 'oracledb'
    jdbc_class = 'oracle.jdbc.driver.OracleDriver'

    @classmethod
    def connect(cls, conf: ParamBunch, *args,
                user=None, password=None, host=None, port=None,
                schema=None, sid=None, service_name=None,
                dsn=None, **kwargs):
        """
        use the appropriate underlying library to connect to the database
        """
        from oracledb import makedsn, connect as oci_connect  # pylint: disable=E0611
        user = user or conf.get('user')
        password = password or conf.get('password')
        host = host or conf.get('host')
        port = port or conf.get('port') or cls.default_port
        service_name = service_name or conf.get('service_name')
        sid = sid or conf.get('sid')
        dsn = dsn or conf.get('dsn')
        if not dsn:
            if service_name and not sid:
                dsn = makedsn(host, port, service_name=service_name)
            else:
                dsn = makedsn(host, port, sid)
        connection = oci_connect(user=user, password=password, dsn=dsn, **kwargs)
        schema = schema or conf.get('schema')
        if schema:
            with connection.cursor() as cursor:
                cursor.execute(
                    f'alter session set current_schema={schema}'
                    "  nls_date_format='yyyy-mm-dd hh24:mi:ss'"
                    "  nls_timestamp_format='yyyy-mm-dd hh24:mi:ss'"
                )
        else:
            with connection.cursor() as cursor:
                cursor.execute(
                    'alter session set'
                    "  nls_date_format='yyyy-mm-dd hh24:mi:ss'"
                    "  nls_timestamp_format='yyyy-mm-dd hh24:mi:ss'"
                )
        return connection

    @classmethod
    def sql_alchemy_connection_string(
            cls, conf: ParamBunch, *args,
            host=None, port=None, sid=None, service_name=None,
            dsn=None, **kwargs):
        """
        the sql alchemy connection string used to connect to the database
        """
        from cx_Oracle import makedsn
        host = host or conf.get('host')
        port = port or conf.get('port') or cls.default_port
        service_name = conf.get('service_name')
        sid = sid or conf.get('sid')
        dsn = dsn or conf.get('dsn')
        if not dsn:
            if service_name and not dsn:
                dsn = makedsn(host, port, service_name=service_name)
            else:
                dsn = makedsn(host, port, sid)
        return super().sql_alchemy_connection_string(
            conf, suffix=dsn, **kwargs)

    @classmethod
    def jdbc_connection_string(cls, conf: ParamBunch, *args,
                               host=None, port=None, sid=None,
                               service_name=None, dsn=None, **kwargs):
        """
        the jdbc connection string used to connect to the database
        """
        from cx_Oracle import makedsn
        host = host or conf.get('host')
        port = port or conf.get('port') or cls.default_port
        service_name = conf.get('service_name')
        sid = sid or conf.get('sid')
        dsn = dsn or conf.get('dsn')
        if service_name and not dsn:
            dsn = makedsn(conf.host, port, service_name=service_name)
        if dsn:
            return f'jdbc:oracle:thin:@{dsn}'
        return f'jdbc:oracle:thin:@{host}:{port}/{sid}'



from waddle import ParamBunch

from .odbc import OdbcConnector


class Hp3000Odbc(OdbcConnector):
    jdbc_class = 'com.minisoft.jdbc.MSJDBCDriver'
    default_port = 32233
    default_driver = 'hp3000'

    @classmethod
    def connect(cls, conf: ParamBunch, *args, **kwargs):
        import pyodbc
        cls.ensure('driver', kwargs, conf)
        driver = kwargs['driver']
        server = kwargs.get('server') or conf.host
        port = kwargs.get('port') or conf.port
        user = kwargs.get('user') or conf.user
        account = kwargs.get('account') or conf.password
        pieces = [
            f'server={server}',
            f'server port={port}',
            f'user={user}',
            f'account={account}',
            f'driver={driver}',
        ]
        schemae = kwargs.get('schema') or conf.schema or []
        for i, x in enumerate(schemae, 1):
            pieces.append(f'schema{i}={x}.SCHEMA.{account}')
        st = ';'.join(pieces)
        print(st)
        connection = pyodbc.connect(st)
        return connection




import atexit
import threading
import logging
from contextlib import contextmanager

from waddle import ParamBunch
from .db2i import Db2i
from .db2i_native import Db2iNative
from .hp3000 import Hp3000
from .hp3000_odbc import Hp3000Odbc
from .mssql import SqlServer
from .mysql import Mysql
from .netsuite import Netsuite
from .oracle import Oracle
from .oracle_db import OracleDb
from .postgres import Postgres
from .redshift import Redshift
from .snowflake import Snowflake
from .vertica import Vertica


gtl = threading.local()
log = logging.getLogger(__name__)



class ConnectionManager:
    thread_local_key = 'concentric_connection_manager'


    def __init__(self):
        self.conf = ParamBunch()

    def add_config(self, *filenames):
        for filename in filenames:
            self.conf.from_file(filename, True)

    @classmethod
    def get_instance(cls, *filenames):
        global gtl
        manager = getattr(gtl, cls.thread_local_key, None)
        if manager:
            return manager
        manager = cls()
        manager.add_config(*filenames)
        setattr(gtl, cls.thread_local_key, manager)
        return manager

    @classmethod
    def get_connector(cls, alias):
        manager = cls.get_instance()
        conf = manager.conf[alias]
        if conf.engine == 'oracle':
            return conf, Oracle
        if conf.engine == 'oracledb':
            return conf, OracleDb
        if conf.engine == 'db2i':
            return conf, Db2i
        if conf.engine == 'db2i_native':
            return conf, Db2iNative
        if conf.engine == 'hp3000':
            return conf, Hp3000
        if conf.engine == 'hp3000_odbc':
            return conf, Hp3000Odbc
        if conf.engine == 'mssql':
            return conf, SqlServer
        if conf.engine == 'mysql':
            return conf, Mysql
        if conf.engine == 'netsuite':
            return conf, Netsuite
        if conf.engine == 'postgres':
            return conf, Postgres
        if conf.engine == 'redshift':
            return conf, Redshift
        if conf.engine == 'snowflake':
            return conf, Snowflake
        if conf.engine == 'vertica':
            return conf, Vertica
        return None, None

    @classmethod
    def connect(cls, alias, ping=False):
        conf, connector = cls.get_connector(alias)
        return connector.connect(conf)


class CachingConnectionManager(ConnectionManager):
    thread_local_key = 'concentric_caching_connection_manager'
    thread_local_cache_key = 'concentric_connection_cache'

    @classmethod
    def test_connection(cls, alias, conn):
        _, engine = cls.get_connector(alias)
        ping_query = getattr(engine, 'ping_query')
        if ping_query:
            try:
                with conn.cursor() as cursor:
                    cursor.execute(ping_query)
                    cursor.fetchall()
            except:
                conn = None
        return conn

    @classmethod
    def connect(cls, alias, use_cache=True, pre_ping=True):
        global gtl
        cache = getattr(gtl, cls.thread_local_cache_key, None) or {}
        if use_cache:
            conn = cache.get(alias)
            if pre_ping and conn:
                if not hasattr(conn, 'ping'):
                    conn = cls.test_connection(alias, conn)
                else:
                    try:
                        conn.ping()
                    except:
                        conn = None
            if conn:
                return conn
        conn = super().connect(alias)
        if use_cache:
            cache[alias] = conn
            setattr(gtl, cls.thread_local_cache_key, cache)
        return conn

    @classmethod
    def cleanup(cls):
        global gtl
        cache = getattr(gtl, cls.thread_local_cache_key, None) or {}
        keys = []
        for alias, conn in cache.items():
            try:
                keys.append(alias)
                conn.close()
            except:
                pass
        for alias in keys:
            cache.pop(alias, None)


@contextmanager
def transactional(session_cls, initial):
    """
    Provide a transactional scope around a series of operations.
    """
    session = session_cls()
    try:
        for x in initial:
            session.execute(x)
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


@contextmanager
def read_only(session_cls, initial, **params):
    """
    Provide a closing scope around a series of operations.
    used for query operations
    """
    session = session_cls(**params)
    try:
        for x in initial:
            session.execute(x)
        yield session
    finally:
        session.close()


class Alchemist:
    """

    Provides thread safe named access to sqlalchemy sessions

    first declare your connection::

        ConnectionMgr.connection("default", "sqlite:///simple.db")

    then connect::

        with ConnectionMgr.session("default") as session:
            session.execute(sql.insert(role, {"name": "user"}))
            session.commit()

    """

    connections = {}
    connections_lock = threading.Lock()

    @classmethod
    def connection(cls, using, force=False, initial=None, **kwargs):
        """
        creates a cached engine and session_cls for db_url
        kwargs are passed on the create_engine
        """
        from sqlalchemy import create_engine
        from sqlalchemy.orm import sessionmaker
        with cls.connections_lock:
            conn = cls.connections.get(using)
            if conn and not force:
                return conn
            conf, connector = ConnectionManager.get_connector(using)
            url = connector.sql_alchemy_connection_string(conf)
            initial = initial or conf.get('initial') or []
            kwargs.setdefault('pool_pre_ping', True)
            engine = create_engine(url, **kwargs)
            session_cls = sessionmaker(bind=engine)
            conn = engine, session_cls, initial
            cls.connections[using] = conn
        return conn

    @classmethod
    def transactional(cls, using, force=False):
        """
        returns a sql alchemy session for the provided
        database connection that should be defined in waddle
        use the ``force`` flag to force this function to return
        a new connection instead of using the cached connection.
        The semantics of a transactional session are such that at the beginning
        of the context, we will begin a transaction, and once the context
        exits, we will either commit or rollback the transaction.
        """
        conn = cls.connection(using, force=force)
        return transactional(conn[1], conn[2])

    @classmethod
    def read_only(cls, name='default', force=False, **params):
        """
        returns a sql alchemy session for the provided
        database connection that should be defined in waddle
        use the ``force`` flag to force this function to return
        a new connection instead of using the cached connection.
        The semantics of a read only session are that no transactional
        management will occur in the session context.
        """
        conn = cls.connection(name, force=force, **params)
        return read_only(conn[1], conn[2])


def cleanup_cached_connections():
    m = CachingConnectionManager.get_instance()
    m.cleanup()


atexit.register(cleanup_cached_connections)



import logging

from waddle import ParamBunch

from .odbc import OdbcConnector


log = logging.getLogger(__name__)


class ProgressOpenedge(OdbcConnector):
    default_driver = 'progress_openedge'
    default_port = 8000
    engine = 'progress_openedge'

    @classmethod
    def connect(cls, conf: ParamBunch, *args, name=None,
                host=None, user=None, password=None,
                port=None, **kwargs):
        cls.ensure('driver', kwargs, conf)
        port = port or conf.port or cls.default_port
        user = user or conf.user
        host = host or conf.host
        password = password or conf.password
        kwargs['portnumber'] = port
        kwargs['databasename'] = name or conf.name
        kwargs['uid'] = user
        kwargs['pwd'] = password
        kwargs['hostname'] = host
        return super().connect(conf, **kwargs)



import urllib
import urllib.parse
from waddle import ParamBunch

from .base import BaseConnector


class Db2iNative(BaseConnector):
    engine = 'db2+ibm_db_dbi'
    jdbc_class = 'com.ibm.as400.access.AS400JDBCDriver'
    ping_query = 'select current date from sysibm.sysdummy1'

    @classmethod
    def connect(cls, conf: ParamBunch, **kwargs):
        """
        use the appropriate underlying library to connect to the database
        """
        import ibm_db_dbi as db2
        return db2.connect(conn_options=kwargs)

    @classmethod
    def sql_alchemy_connection_string(
            cls, conf: ParamBunch, *args, host=None, **kwargs):
        """
        the sql alchemy connection string used to connect to the database
        """
        return f'{cls.engine}://*LOCAL'

    @classmethod
    def jdbc_connection_string(cls, conf: ParamBunch, *args, **kwargs):
        """
        the jdbc connection string used to connect to the database
        """
        return f'jdbc:as400://{conf.host}'
